/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/


#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <jni.h>


#include "cadef.h"
#include "db_access.h"
#include "jca_Ca.h"
#include "jca_PV.h"


#define CA_DEBUG 0


#ifdef CA_DEBUG
#define CA_DEBUG0(l,t)       if(l<=CA_DEBUG) printf(t);
#define CA_DEBUG1(l,t,p1)    if(l<=CA_DEBUG) printf(t,p1);
#define CA_DEBUG2(l,t,p1,p2) if(l<=CA_DEBUG) printf(t,p1,p2);
#else
#define CA_DEBUG0(l,t)
#define CA_DEBUG1(l,t,p1)
#define CA_DEBUG2(l,t,p1,p2)
#endif


#define NBASEINC 100

struct jdbr {
  jobject   jdbr;
  jmethodID setInfoMID;
  jfieldID  valueFID;
};
typedef struct jdbr JDBR;

struct jevent {
  jobject   jevent;
};
typedef struct jevent JEVENT;


struct event {
  JDBR    jdbr;
  
  JEVENT  jevent;
  evid    cevid;
};
typedef struct event EVENT;


struct pv {
  chid     chid;

  jobject  jpv;

  JEVENT    jconnection_event;

  EVENT**  plistEvent;
  jint     nbEvent;
};
typedef struct pv PV;

struct get {
  jint    type;
  jint    count;
  JDBR    jdbr;
  void*   cdbr;
};
typedef struct get GET;


static JNIEnv* ENV=0;           /* The JNI environement (unique since only one thread
                                 * is calling those functions                         
                                 *----------------------------------------------------*/
static jclass caClass;           /* The Ca Class
                                 *----------------------------------------------------*/
static jmethodID queueEventID;  /* The method ID for the queueEvent function           
                                 *----------------------------------------------------*/
static PV**    plistPV=0;       /* The array containing all the connected pv
                                 *----------------------------------------------------*/
static jint    nbPV=0;          /* The size of the pv array
                                 *----------------------------------------------------*/

static GET**   plistGet=0;      /* The array containing all the gets DBR
                                 *----------------------------------------------------*/
static jint    maxGet=0;        /* The max number of get DBR
                                 *----------------------------------------------------*/
static jint    nbGet=0;         /* The actual number of get DBR
                                 *----------------------------------------------------*/



/****************************************************************/
/*                           PROTOPYTES                         */
/****************************************************************/


/* set the id of a java PV
 *----------------------------------------------------*/
static void set_id(jobject obj, jint id);


/* Initialize a JDBR struct
 *----------------------------------------------------*/
static void init_JDBR(jint type, JDBR* JDBR, jobject jdbr);

/* Initialize a JEVENT struct
 *----------------------------------------------------*/
static void init_JEVENT(JEVENT* pevent, jobject event);

/* Release a JEVENT struct
 *----------------------------------------------------*/
static void release_JEVENT(JEVENT* pevent);

/* remove a pv from the PV list
 *----------------------------------------------------*/
static void remove_pv(PV** ppv);

/* remove all pvs from the PV list
 *----------------------------------------------------*/
static void remove_pvs();

/* pend all outstanding gets
 *----------------------------------------------------*/
static void pend_gets();

/* Create a cdbr buffer of type type and count count
 *----------------------------------------------------*/
static void* create_cdbr(jint type, jint count);

/* Copy the values of a cdbr to a jdbr
 *----------------------------------------------------*/
static void cdbr2JDBR(void* cdbr, jint type, jint count, JDBR* jdbr);

/* Copy the values of a jdbr to a cdbr
 *----------------------------------------------------*/
static void JDBR2cdbr(JDBR* jdbr, jint type, jint count, void* cdbr);

/* Create a GET
 *----------------------------------------------------*/
static GET* create_GET(jint type, jint count, jobject jdbr);

/* Destroy a GET
 *----------------------------------------------------*/
static void destroy_GET(GET** pget);




/****************************************************************/
/*                         IMPLEMENTATION                       */
/****************************************************************/

/* set the id field of an object
 * The java ID field must be named m_id and must be of type int.
 *----------------------------------------------------*/
static void set_id(jobject obj, jint id) {
  jclass   Class= (*ENV)->GetObjectClass(ENV, obj);
  jfieldID fID= (*ENV)->GetFieldID(ENV,Class,"m_id","I");
  (*ENV)->SetIntField(ENV,obj,fID,id);
}

/* Initialize a JDBR struct
 *----------------------------------------------------*/
static void init_JDBR(jint type, JDBR* pjdbr, jobject jdbr) {
  jclass jClass;

  pjdbr->jdbr= jdbr;
  jClass= (*ENV)->GetObjectClass(ENV,jdbr);

  pjdbr->setInfoMID= NULL;
  pjdbr->valueFID= NULL;

	if(dbr_type_is_STRING(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "Ljava/lang/String;");
	} else if(dbr_type_is_ENUM(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[S");
	} else if(dbr_type_is_SHORT(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[S");
	} else if(dbr_type_is_FLOAT(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[F");
	} else if(dbr_type_is_CHAR(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[B");
  } else if(dbr_type_is_LONG(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[I");
	} else if(dbr_type_is_DOUBLE(type)) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "[D");
	} else if(type==DBR_PUT_ACKS) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "S");
  } else if(type==DBR_PUT_ACKT) {
    pjdbr->valueFID= (*ENV)->GetFieldID(ENV,jClass, "m_value", "S");
  }

	if(type==DBR_STSACK_STRING) {
    pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass, "setInfo",
                                           "(SSSS)V");
	} else if(dbr_type_is_STS(type)) {
   pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass, "setInfo", "(SS)V");
	} else if(dbr_type_is_TIME(type)) {
   pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass, "setInfo", "(SSII)V");
  } else {
		switch(type) {
      case DBR_GR_ENUM:  /* this is really the java DBR.INFO_ENUM type */
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SS[Ljava/lang/String;)V");
				break;

			case DBR_GR_SHORT:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;SSSSSS)V");
				break;
			case DBR_GR_FLOAT:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSSLjava/lang/String;FFFFFF)V");
				break;
			case DBR_GR_CHAR:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;BBBBBB)V");
				break;
			case DBR_GR_LONG:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;IIIIII)V");
				break;
			case DBR_GR_DOUBLE:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSSLjava/lang/String;DDDDDD)V");
				break;

			case DBR_CTRL_SHORT:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;SSSSSSSS)V");
				break;
			case DBR_CTRL_FLOAT:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSSLjava/lang/String;FFFFFFFF)V");
				break;
			case DBR_CTRL_CHAR:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;BBBBBBBB)V");
				break;
			case DBR_CTRL_LONG:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSLjava/lang/String;IIIIIIII)V");
				break;
			case DBR_CTRL_DOUBLE:
				pjdbr->setInfoMID= (*ENV)->GetMethodID(ENV,jClass,"setInfo","(SSSLjava/lang/String;DDDDDDDD)V");
				break;
		}
  }
}

static void init_JEVENT(JEVENT* pevent, jobject jevent) {
  if(jevent!=NULL) {
    pevent->jevent= (*ENV)->NewGlobalRef(ENV,jevent);
  } else {
    pevent->jevent= NULL;
  }
}

static void release_JEVENT(JEVENT* pevent) {
  if(pevent->jevent!=NULL) {
    (*ENV)->DeleteGlobalRef(ENV,pevent->jevent);
  }
}

static EVENT* create_event(jint type, jobject jdbr, jobject jevent) {
  EVENT* event= (EVENT*) malloc(sizeof(EVENT));
  
  if(jdbr!=NULL) {
    init_JDBR(type, &event->jdbr, (*ENV)->NewGlobalRef(ENV,jdbr));
  } else {
    event->jdbr.jdbr= NULL;
  }

  init_JEVENT(&event->jevent, jevent);

  return event;
}


static void destroy_event(EVENT** pevent) {

  if((*pevent)->jdbr.jdbr!=NULL) {
    (*ENV)->DeleteGlobalRef(ENV, (*pevent)->jdbr.jdbr);
  }

  release_JEVENT(&(*pevent)->jevent);

  free(*pevent);
  *pevent= NULL;
}

static EVENT** add_event(PV* pv,
                         jint type, jobject jdbr,
                         jobject jevent) {
  jint t, id;
  jint nbEvent= pv->nbEvent;
  EVENT** plistEvent= pv->plistEvent;

  id= -1;
  for(t=0; t<nbEvent; ++t) {
    if(plistEvent[t]==NULL) {
      id= t;
      break;
    }
  }
  if(id==-1) {
    id= nbEvent;
    nbEvent+=NBASEINC;
    plistEvent= (EVENT**) malloc(nbEvent*sizeof(EVENT*));
    for(t=0; t<pv->nbEvent; ++t) plistEvent[t]= pv->plistEvent[t];
    for(t=pv->nbEvent; t<nbEvent; ++t) plistEvent[t]= NULL;
    free(pv->plistEvent);
    pv->plistEvent= plistEvent;
    pv->nbEvent= nbEvent;
  }

  plistEvent[id]= create_event(type, jdbr, jevent);

  return &plistEvent[id];
}  
                           
/* remove a pv from the PV list
 *----------------------------------------------------*/
static void remove_pv(PV** ppv) {
  jint t;

  set_id((*ppv)->jpv, -1);
  (*ENV)->DeleteGlobalRef(ENV,(*ppv)->jpv);

  release_JEVENT(&(*ppv)->jconnection_event);

  for(t= 0; t<(*ppv)->nbEvent; ++t) {
    if((*ppv)->plistEvent[t]!=NULL) destroy_event(&(*ppv)->plistEvent[t]);
  }
  free((*ppv)->plistEvent);
  free((*ppv));
  *ppv= NULL;
}

/* remove all pvs from the PV list
 *----------------------------------------------------*/
static void remove_pvs() {
  jint t;
  for(t=0; t<nbPV; ++t) {
    if(plistPV[t]!=NULL) remove_pv(&plistPV[t]);
  }
}

/* pend all outstanding gets
 *----------------------------------------------------*/
static void pend_gets() {
  int t;

  for(t=0; t<nbGet; ++t) {
    cdbr2JDBR(plistGet[t]->cdbr, plistGet[t]->type, 
              plistGet[t]->count, &plistGet[t]->jdbr);
    destroy_GET(&plistGet[t]);
  }
  nbGet= 0;
}


/* Create a cdbr buffer of type type and count count
 *----------------------------------------------------*/
static void* create_cdbr(jint type, jint count) {
  long size;

  size= dbr_size[type]+(count-1)*dbr_value_size[type];
  return malloc(size);
}

/* Copy the values of a cdbr to a jdbr
 *----------------------------------------------------*/
static void cdbr2JDBR(void* cdbr, jint type, jint count, JDBR* jdbr) {
	void* ptr;
	CA_DEBUG1(5,"cdbr2JDBR %d\n", type);

	ptr=  dbr_value_ptr(cdbr, type);
	if(dbr_type_is_STRING(type)) {
    jstring jstr= (*ENV)->NewStringUTF(ENV, (char*)ptr);
    (*ENV)->SetObjectField(ENV, jdbr->jdbr, jdbr->valueFID, jstr);
  } else if(dbr_type_is_SHORT(type)) {
    jshortArray array= (jshortArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetShortArrayRegion(ENV, array, 0, count, (jshort*)ptr);
  } else if(dbr_type_is_ENUM(type)) {
    jshortArray array= (jshortArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetShortArrayRegion(ENV, array, 0, count, (jshort*)ptr);
  } else if(dbr_type_is_FLOAT(type)) {
    jfloatArray array= (jfloatArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetFloatArrayRegion(ENV, array, 0, count, (jfloat*)ptr);
  } else if(dbr_type_is_CHAR(type)) {
    jbyteArray array= (jbyteArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetByteArrayRegion(ENV, array, 0, count, (jbyte*)ptr);
  } else if(dbr_type_is_LONG(type)) {
    jintArray array= (jintArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetIntArrayRegion(ENV, array, 0, count, (jint*)ptr);
  } else if(dbr_type_is_DOUBLE(type)) {
    jdoubleArray array= (jdoubleArray) (*ENV)->GetObjectField(ENV, jdbr->jdbr, jdbr->valueFID);
    (*ENV)->SetDoubleArrayRegion(ENV, array, 0, count, (jdouble*)ptr);
  }

	switch(type) {
		case DBR_STS_STRING: {
			struct dbr_sts_string* ptr= (struct dbr_sts_string*)cdbr;
			(*ENV)->CallVoidMethod(ENV,jdbr->jdbr, jdbr->setInfoMID,
			                       ptr->status, ptr->severity);
			break;
		}
  case DBR_STS_SHORT: {
    struct dbr_sts_short* ptr= (struct dbr_sts_short*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }
  case DBR_STS_ENUM: {
    struct dbr_sts_enum* ptr= (struct dbr_sts_enum*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }
  case DBR_STS_FLOAT: {
    struct dbr_sts_float* ptr= (struct dbr_sts_float*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }
  case DBR_STS_CHAR: {
    struct dbr_sts_char* ptr= (struct dbr_sts_char*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }
  case DBR_STS_LONG: {
    struct dbr_sts_long* ptr= (struct dbr_sts_long*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }
  case DBR_STS_DOUBLE: {
    struct dbr_sts_double* ptr= (struct dbr_sts_double*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity);
    break;
  }

		case DBR_TIME_STRING: {
			struct dbr_time_string* ptr= (struct dbr_time_string*)cdbr;
			(*ENV)->CallVoidMethod(ENV,jdbr->jdbr, jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
			break;
		}
  case DBR_TIME_SHORT: {
    struct dbr_time_short* ptr= (struct dbr_time_short*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }
  case DBR_TIME_ENUM: {
    struct dbr_time_enum* ptr= (struct dbr_time_enum*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }
  case DBR_TIME_FLOAT: {
    struct dbr_time_float* ptr= (struct dbr_time_float*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }
  case DBR_TIME_CHAR: {
    struct dbr_time_char* ptr= (struct dbr_time_char*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }
  case DBR_TIME_LONG: {
    struct dbr_time_long* ptr= (struct dbr_time_long*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }
  case DBR_TIME_DOUBLE: {
    struct dbr_time_double* ptr= (struct dbr_time_double*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
			                       ptr->status, ptr->severity, ptr->stamp.secPastEpoch, ptr->stamp.nsec);
    break;
  }



  case DBR_GR_ENUM: { /* this is really the java DBR.INFO_ENUM type */
    int t;
    char buffer[MAX_ENUM_STRING_SIZE+1];
    jstring jstr;
    struct dbr_gr_enum* ptr= (struct dbr_gr_enum*)cdbr;
    jarray values= (*ENV)->NewObjectArray(ENV, ptr->no_str, 
                                          (*ENV)->FindClass(ENV,"java/lang/String"), NULL);
    if(values!=NULL) {
      for(t=0; t<ptr->no_str; ++t) {
        strncpy(buffer, ptr->strs[t], MAX_ENUM_STRING_SIZE);
        buffer[MAX_ENUM_STRING_SIZE]='\0';
        jstr= (*ENV)->NewStringUTF(ENV, buffer);
        (*ENV)->SetObjectArrayElement(ENV, values, t, jstr);
      }
    }

		(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity, values);
    break;
  }

  case DBR_GR_SHORT: {
    struct dbr_gr_short* ptr= (struct dbr_gr_short*)cdbr;
		(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit);
    break;
  }
  case DBR_GR_FLOAT: {
    struct dbr_gr_float* ptr= (struct dbr_gr_float*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity, ptr->precision,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit);
    break;
  }
  case DBR_GR_CHAR: {
    struct dbr_gr_char* ptr= (struct dbr_gr_char*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit);
    break;
  }
  case DBR_GR_LONG: {
    struct dbr_gr_long* ptr= (struct dbr_gr_long*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit);
    break;
  }
  case DBR_GR_DOUBLE: {
    struct dbr_gr_double* ptr= (struct dbr_gr_double*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity, ptr->precision,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit);
    break;
  }
  case DBR_CTRL_SHORT: {
    struct dbr_ctrl_short* ptr= (struct dbr_ctrl_short*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit,
						   ptr->upper_ctrl_limit,
						   ptr->lower_ctrl_limit);
    break;
  }
  case DBR_CTRL_FLOAT: {
    struct dbr_ctrl_float* ptr= (struct dbr_ctrl_float*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity, ptr->precision,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit,
						   ptr->upper_ctrl_limit,
						   ptr->lower_ctrl_limit);
    break;
  }
  case DBR_CTRL_CHAR: {
    struct dbr_ctrl_char* ptr= (struct dbr_ctrl_char*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit,
						   ptr->upper_ctrl_limit,
						   ptr->lower_ctrl_limit);
    break;
  }
  case DBR_CTRL_LONG: {
    struct dbr_ctrl_long* ptr= (struct dbr_ctrl_long*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit,
						   ptr->upper_ctrl_limit,
						   ptr->lower_ctrl_limit);
    break;
  }
  case DBR_CTRL_DOUBLE: {
    struct dbr_ctrl_double* ptr= (struct dbr_ctrl_double*)cdbr;
	(*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status, ptr->severity, ptr->precision,
						   (*ENV)->NewStringUTF(ENV, ptr->units),
						   ptr->upper_disp_limit,
						   ptr->lower_disp_limit,
						   ptr->upper_alarm_limit,
						   ptr->upper_warning_limit,
						   ptr->lower_warning_limit,
						   ptr->lower_alarm_limit,
						   ptr->upper_ctrl_limit,
						   ptr->lower_ctrl_limit);
    break;
  }
  case DBR_STSACK_STRING: {
    struct dbr_stsack_string* ptr= (struct dbr_stsack_string*)cdbr;
    (*ENV)->CallVoidMethod(ENV,jdbr->jdbr,jdbr->setInfoMID,
                           ptr->status,ptr->severity,ptr->ackt,ptr->acks);
    break;
  }
  case DBR_PUT_ACKS: {
    (*ENV)->SetShortField(ENV,jdbr->jdbr,jdbr->valueFID,*((jshort*)cdbr));
    break;
  }
  case DBR_PUT_ACKT: {
    (*ENV)->SetShortField(ENV,jdbr->jdbr,jdbr->valueFID,*((jshort*)cdbr));
    break;
  }
  }
}

/* Copy the values of a jdbr to a cdbr
 *----------------------------------------------------*/
static void JDBR2cdbr(JDBR* jdbr, jint type, jint count, void* cdbr) {

  switch(type) {
  case DBR_STRING: {
    jstring     jstr;
    const char* cstr;
    jstr= (jstring) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    
    cstr= (*ENV)->GetStringUTFChars(ENV, jstr, NULL);
    strncpy((char*)cdbr, cstr, 39);
    ((char*)cdbr)[39]= '\0';
    (*ENV)->ReleaseStringUTFChars(ENV, jstr, cstr);
    break;
  }
  case DBR_SHORT: {
    jshortArray array;
    array= (jshortArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetShortArrayRegion(ENV, array, 0,count,(jshort*)cdbr);
    break;
  }
  case DBR_ENUM: {
    jshortArray array;
    array= (jshortArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetShortArrayRegion(ENV, array, 0,count,(jshort*)cdbr);
    break;
  }
  case DBR_FLOAT: {
    jfloatArray array;
    array= (jfloatArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetFloatArrayRegion(ENV, array, 0,count,(jfloat*)cdbr);
    break;
  }
  case DBR_CHAR: {
    jbyteArray array;
    array= (jbyteArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetByteArrayRegion(ENV, array, 0,count,(jbyte*)cdbr);
    break;
  }
  case DBR_LONG: {
    jintArray array;
    array= (jintArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetIntArrayRegion(ENV, array, 0,count,(jint*)cdbr);
    break;
  }
  case DBR_DOUBLE: {
    jdoubleArray array;
    array= (jdoubleArray) (*ENV)->GetObjectField(ENV,jdbr->jdbr,jdbr->valueFID);
    (*ENV)->GetDoubleArrayRegion(ENV, array, 0,count,(jdouble*)cdbr);
    break;
  }
  }
}

/* Create a GET
 *----------------------------------------------------*/
static GET* create_GET(jint type, jint count, jobject jdbr) {
  GET* get= (GET*) malloc(sizeof(GET));

  get->type= type;
  get->count= count;
  init_JDBR(type,&get->jdbr, (*ENV)->NewGlobalRef(ENV,jdbr));
  get->cdbr= create_cdbr(type,count);
  
  return get;
}

/* Destroy a CGET
 *----------------------------------------------------*/
static void destroy_GET(GET** pget) {
  (*ENV)->DeleteGlobalRef(ENV, (*pget)->jdbr.jdbr);
  free((*pget)->cdbr);
  free((*pget));
  *pget= NULL;
}

          
/****************************************************************
 * JNI implementation !!!
 ****************************************************************/

static void my_callback(struct event_handler_args ARGS) {
  printf("my_callback...\n");
  printf("  type     : %d\n", ARGS.type);
  printf("  count    : %d\n", ARGS.count);
  printf("  dbr      : %X\n", ARGS.dbr);
  printf("  chid     : %X\n", ARGS.chid);
  printf("  status   : %d\n", ARGS.status);
  printf("...my_callback\n");
}


static void connection_callback(struct connection_handler_args ARGS) {
  int t;
  int id= -1;

  CA_DEBUG0(2, "connection_callback...\n");

  for(t=0; t<nbPV; ++t) {
    if(plistPV[t]!=NULL && plistPV[t]->chid==ARGS.chid) {
      id= t;
      break;
    }
  }
  if(id==-1) {
    /* error */
    return;
  }
  (*ENV)->CallStaticVoidMethod(ENV, caClass, queueEventID,
                               plistPV[id]->jconnection_event.jevent,
                               ARGS.op);

  CA_DEBUG0(2, "...connection_callback\n");
}  

static void get_callback(struct event_handler_args ARGS) {
  EVENT** pevent= (EVENT**) ARGS.usr;

  CA_DEBUG0(2, "get_callback...\n");

  cdbr2JDBR((void*)ARGS.dbr,ARGS.type,ARGS.count,&(*pevent)->jdbr);

  (*ENV)->CallStaticVoidMethod(ENV, caClass, queueEventID,
                               (*pevent)->jevent.jevent,
                               ARGS.status);

  destroy_event(pevent);
  CA_DEBUG0(2, "...get_callback\n");
}


static void put_callback(struct event_handler_args ARGS) {
  EVENT** pevent= (EVENT**) ARGS.usr;

  CA_DEBUG0(2, "put_callback...\n");
  (*ENV)->CallStaticVoidMethod(ENV, caClass, queueEventID,
                               (*pevent)->jevent.jevent,
                               ARGS.status);

  destroy_event(pevent);
  CA_DEBUG0(2, "...put_callback\n");
}

static void event_callback(struct event_handler_args ARGS) {
  EVENT** pevent= (EVENT**) ARGS.usr;

  CA_DEBUG0(2, "event_callback...\n");

  cdbr2JDBR((void*)ARGS.dbr,ARGS.type,ARGS.count,&(*pevent)->jdbr);

  (*ENV)->CallStaticVoidMethod(ENV, caClass, queueEventID,
                               (*pevent)->jevent.jevent,
                               ARGS.status);
  CA_DEBUG0(2, "...event_callback\n");
}




/* jni_init(); */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1init(JNIEnv * env,
                                             jclass jca) {
  jint status;
  CA_DEBUG0(1, "init...\n");

  if(ENV!=0) return ECA_NORMAL;
  ENV= env;
  caClass= (jclass) (*ENV)->NewGlobalRef(ENV, jca);
  queueEventID= (*ENV)->GetStaticMethodID(ENV, caClass, "queueEvent", 
                                          "(Ljca/event/CaEvent;I)V");
  status= ca_task_initialize();
  CA_DEBUG0(1, "...init\n");
  return status;
}


/* jni_exit(); */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1exit(JNIEnv *env,
                                             jclass jca) {
  jint status;
  if(ENV==0) return ECA_NORMAL;
  (*ENV)->DeleteGlobalRef(ENV, caClass);
  CA_DEBUG0(1, "exit...\n");
  status= ca_task_exit();

  remove_pvs();
  free(plistPV);
  plistPV= 0;
  nbPV= 0;

  pend_gets();
  free(plistGet);
  plistGet= 0;
  maxGet= 0;
  nbGet= 0;

  ENV= 0;
  CA_DEBUG0(1, "...exit\n");
  return status;
}

/* jni_search() */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1search(JNIEnv *env, 
                                               jclass jca,
                                               jobject jpv,
                                               jobject jconnection_event) {
  jint        status;
  jclass      jClass;
  jfieldID    jfID;
  jstring     jstr;
  const char* cstr;
  PV*         pv;
  jint        id;
  jint        t;

  /* Determine the PV id */
  id= -1;
  for(t=0; t<nbPV; ++t) {
    if(plistPV[t]==NULL) {
      id= t;
      break;
    }
  }
  if(id==-1) {
    PV** tmp= (PV**) malloc((nbPV+NBASEINC)*sizeof(PV*));
    for(t=0; t<nbPV; ++t) tmp[t]= plistPV[t];
    for(t=0; t<NBASEINC; ++t) tmp[t+nbPV]= NULL;
    free(plistPV);
    plistPV= tmp;
    id= nbPV;
    nbPV+=NBASEINC;
  }
  /* Create the new pv and add it to the list */
  plistPV[id]= pv= (PV*) malloc(sizeof(PV));

  /* Initialize the pv fields */
  pv->jpv= (*ENV)->NewGlobalRef(ENV, jpv);

  init_JEVENT(&pv->jconnection_event, jconnection_event);

  pv->plistEvent= NULL;
  pv->nbEvent= 0;


  /* Connect the new PV */
  jClass= (*ENV)->GetObjectClass(ENV, jpv);
  jfID= (*ENV)->GetFieldID(ENV,jClass, "m_name", "Ljava/lang/String;");
  jstr= (jstring) (*ENV)->GetObjectField(ENV,jpv,jfID);
  cstr= (*ENV)->GetStringUTFChars(ENV,jstr,0);

  if(jconnection_event==NULL) {
    CA_DEBUG1(1, "search %s ...\n",cstr);
    status= ca_search(cstr, &pv->chid);
  } else {
    CA_DEBUG1(1, "search_and_connect %s ...\n",cstr);
    status= ca_search_and_connect(cstr, &pv->chid, connection_callback, NULL);
  }

  (*ENV)->ReleaseStringUTFChars(ENV,jstr,cstr);

  set_id(jpv, id);
  CA_DEBUG0(1, "...search\n");
  return status;
}

/* jni_clear_channel() */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1clear_1channel(JNIEnv *env,
                                                       jclass jca, 
                                                       jint   id) {
  jint status;
  CA_DEBUG0(1, "clear...\n");
  status= ca_clear_channel(plistPV[id]->chid);

  remove_pv(&plistPV[id]);
  CA_DEBUG0(1, "...clear\n");
  return status;
}

/* jni_flush_io() */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1flush_1io(JNIEnv *env,
                                                  jclass jca) {
  jint status;
  CA_DEBUG0(3, "flushIO\n");
  status= ca_flush_io();
  return status;
}

/* jni_pend_event(double timeout) */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1pend_1event(JNIEnv *env,
                                                    jclass jca,
                                                    jdouble timeout) {
  jint status;
  CA_DEBUG0(3, "pendEvent\n");
  status= ca_pend_event(timeout);
  return status;
}

/* jni_pend_io(double timeout) */
JNIEXPORT jint JNICALL Java_jca_Ca_jni_1pend_1io(JNIEnv *env,
                                                 jclass jca,
                                                 jdouble timeout) {

  jint status;
  CA_DEBUG0(3, "pendIO\n");

  status= ca_pend_io((float)timeout);

  pend_gets();
  return status;
}

JNIEXPORT jint JNICALL Java_jca_Ca_jni_1put(JNIEnv *env,
                                            jclass jca,
                                            jint type,
                                            jint count,
                                            jint id,
                                            jobject jdbr) {
  jint status;
  void* cdbr= create_cdbr(type,count);
  JDBR JDBR;

  CA_DEBUG0(4, "put...\n");

  init_JDBR(type, &JDBR, jdbr);
  JDBR2cdbr(&JDBR, type, count, cdbr);
  status= ca_array_put(type,count, plistPV[id]->chid, cdbr);
  free(cdbr);
  CA_DEBUG0(4, "...put\n");
  return status;
}

JNIEXPORT jint JNICALL Java_jca_Ca_jni_1put_1callback(JNIEnv *env,
                                                      jclass jca,
                                                      jint type,
                                                      jint count,
                                                      jint id,
                                                      jobject jdbr,
                                                      jobject jevent) {

  jint     status;
  EVENT**  pevent;
  void*    cdbr;
  JDBR     JDBR;

  CA_DEBUG0(4, "put_callback...\n");

  pevent= add_event(plistPV[id], 0, NULL, jevent);
  cdbr= create_cdbr(type,count);
  init_JDBR(type, &JDBR, jdbr);
  JDBR2cdbr(&JDBR,type,count,cdbr);

  status= ca_array_put_callback(type,count, plistPV[id]->chid, cdbr, put_callback, pevent);
  free(cdbr);
  CA_DEBUG0(4, "...put_callback\n");
  return status;
}


JNIEXPORT jint JNICALL Java_jca_Ca_jni_1get(JNIEnv *env,
                                            jclass jca,
                                            jint type,
                                            jint count,
                                            jint id,
                                            jobject jdbr) {
  jint status;
  jint t,getId;

  CA_DEBUG0(4, "get...\n");
  if(nbGet==maxGet) {
    GET** tmp;
    maxGet+=NBASEINC;
    tmp= (GET**) malloc(maxGet*sizeof(GET*));
    for(t=0; t<nbGet; ++t) {
      tmp[t]= plistGet[t];
    }
    free(plistGet);
    plistGet= tmp;
  }
  getId= nbGet++;
  plistGet[getId]= create_GET(type,count,jdbr);

  status= ca_array_get(type, count, plistPV[id]->chid, plistGet[getId]->cdbr);
  CA_DEBUG0(4, "...get\n");
  return status;
}




JNIEXPORT jint JNICALL Java_jca_Ca_jni_1get_1callback(JNIEnv *env,
                                                      jclass jca,
                                                      jint type,
                                                      jint count,
                                                      jint id,
                                                      jobject jdbr,
                                                      jobject jevent) {
  jint status;
  EVENT** pevent;

  CA_DEBUG0(4, "get_callback...\n");
  pevent= add_event(plistPV[id], type, jdbr, jevent);

  status= ca_array_get_callback(type,count, plistPV[id]->chid, get_callback, pevent);
  CA_DEBUG0(4, "...get_callback\n");
  return status;
}


JNIEXPORT jint JNICALL Java_jca_Ca_jni_1add_1monitor(JNIEnv *env,
                                                     jclass jca,
                                                     jint type,
                                                     jint count,
                                                     jint id,
                                                     jobject jdbr,
                                                     jobject jevent,
                                                     jint mask) {
  jint status;
  EVENT** pevent;

  CA_DEBUG0(4, "add_monitor...\n");

  pevent= add_event(plistPV[id], type, jdbr, jevent);
  set_id(jevent, (jint)pevent);

  status= ca_add_masked_array_event(type,count, plistPV[id]->chid, event_callback, pevent,
                                    (float)0, (float)0, (float)0,
                                    &(*pevent)->cevid, mask);
  CA_DEBUG0(4, "...add_monitor\n");
  return status;
}

JNIEXPORT jint JNICALL Java_jca_Ca_jni_1clear_1monitor(JNIEnv *env,
                                                       jclass jca,
                                                       jint   id) {
  jint status;
  EVENT** pevent= (EVENT**)id;
  CA_DEBUG0(4, "clear_monitor...\n");

  status= ca_clear_event((*pevent)->cevid);
  
  destroy_event(pevent);
  CA_DEBUG0(4, "...clear_monitor\n");
  return status;
}

JNIEXPORT jint JNICALL Java_jca_Ca_jni_1connection_1listener(JNIEnv *env,
                                                             jclass jca,
                                                             jint id,
                                                             jobject jconnection_event) {
  jint status;
  CA_DEBUG0(4, "connection_listener...\n");

  release_JEVENT(&plistPV[id]->jconnection_event);
  init_JEVENT(&plistPV[id]->jconnection_event, jconnection_event);
  
  if(jconnection_event==NULL)
    status= ca_change_connection_event(plistPV[id]->chid,NULL);
  else 
    status= ca_change_connection_event(plistPV[id]->chid,connection_callback);

  CA_DEBUG0(4, "...connection_listener\n");
  return status;
}



JNIEXPORT jint JNICALL Java_jca_PV_jni_1fieldType(JNIEnv *env,
                                                  jclass jchid,
                                                  jint id) {
  return ca_field_type(plistPV[id]->chid);
}

JNIEXPORT jint JNICALL Java_jca_PV_jni_1elementCount(JNIEnv *env,
                                                     jclass jchid,
                                                     jint id) {
  return ca_element_count(plistPV[id]->chid);
}

JNIEXPORT jint JNICALL Java_jca_PV_jni_1state(JNIEnv *env,
                                              jclass jchid,
                                              jint id) {
  return ca_state(plistPV[id]->chid);
}

JNIEXPORT jstring JNICALL Java_jca_PV_jni_1hostName(JNIEnv *env,
                                                    jclass jchid,
                                                    jint id) {
  return (*env)->NewStringUTF(env, ca_host_name(plistPV[id]->chid));
}

JNIEXPORT jboolean JNICALL Java_jca_PV_jni_1readAccess(JNIEnv *env,
                                                       jclass jchid,
                                                       jint id) {
  return (ca_read_access(plistPV[id]->chid)?1:0);
}

JNIEXPORT jboolean JNICALL Java_jca_PV_jni_1writeAccess(JNIEnv *env,
                                                        jclass jchid,
                                                        jint id) {
  return (ca_write_access(plistPV[id]->chid)?1:0);
}




